﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.IO;

namespace Primeval
{
    /// <summary>
    /// This class defines "instructions" for building a bot
    /// </summary>
    class Genome
    {
        /// <summary>
        /// Number of atoms a genome gets on spawn
        /// </summary>
        private const int spawnAtoms = 10;

        /// <summary>
        /// Atoms making up this genome
        /// </summary>
        public List<Atom> atoms;

        /// <summary>
        /// Required atoms that don't take up slots
        /// </summary>
        public List<Atom> mandatoryAtoms;

        /// <summary>
        /// Probability of a particular mutation occuring
        /// </summary>
        private float mutationRate = 0.12f;

        /// <summary>
        /// The length of this genome
        /// </summary>
        public int Length { get { return atoms.Count; } }

        public string name;

        /// <summary>
        /// Create an empty genome
        /// </summary>
        public Genome()
        {
            atoms = new List<Atom>();
            mandatoryAtoms = new List<Atom>();
        }

        /// <summary>
        /// Fill this genome
        /// </summary>
        public Genome Spawn()
        {
            for (int c = 0; c < spawnAtoms; c++)
                AddAtom(new Atom().Spawn());

            //Fill mandatory atoms
            mandatoryAtoms.Add(new Atom().Spawn(AtomType.COLOR));
            mandatoryAtoms.Add(new Atom().Spawn(AtomType.AGGRO));
            mandatoryAtoms.Add(new Atom().Spawn(AtomType.JINK));
            mandatoryAtoms.Add(new Atom().Spawn(AtomType.FREQ));
            mandatoryAtoms.Add(new Atom().Spawn(AtomType.AMP));

            return this;
        }

        /// <summary>
        /// Add an atom to this genome
        /// </summary>
        /// <param name="atom"></param>
        public void AddAtom(Atom atom)
        {
            atoms.Add(atom);
        }

        public void AddMandatoryAtom(Atom atom)
        {
            mandatoryAtoms.Add(atom);
        }

        /// <summary>
        /// Mutate this genome
        /// </summary>
        /// <returns></returns>
        public Genome Mutate()
        {
            Genome ret = new Genome();

            for (int ai = 0; ai < atoms.Count; ai++)
            {
                //Replace this atom with a new one
                if (Game1.random.NextDouble() <= mutationRate)
                    ret.AddAtom(new Atom().Spawn());
                //Keep the atom but maybe mutate it
                else
                    ret.AddAtom(atoms[ai].Mutate());
            }

            //Add an entirely new atom
            if (Game1.random.NextDouble() <= mutationRate)
                ret.AddAtom(new Atom().Spawn());

            //Mandatory atoms
            for (int c = 0; c < mandatoryAtoms.Count; c++)
            {
                //Replace this atom with a new one
                if (Game1.random.NextDouble() <= mutationRate)
                    ret.AddMandatoryAtom(new Atom().Spawn(mandatoryAtoms[c].type));
                //Keep the atom but maybe mutate it
                else
                    ret.AddMandatoryAtom(mandatoryAtoms[c].Mutate());
            }

            return ret;
        }

        /// <summary>
        /// Crossover this genome with another
        /// </summary>
        /// <param name="other"></param>
        /// <returns></returns>
        public Genome Cross(Genome other)
        {
            Genome ret = new Genome();

            for (int c = 0; c < Math.Max(other.Length, Length); c++)
            {
                if (Game1.random.Next(2) == 0)
                {
                    if (c < other.Length) ret.AddAtom(other.atoms[c]);
                }
                else
                {
                    if (c < Length) ret.AddAtom(atoms[c]);
                }
            }

            //Mandatory atoms
            for (int c = 0; c < mandatoryAtoms.Count; c++)
            {
                if (Game1.random.Next(2) == 0)
                    ret.AddMandatoryAtom(other.mandatoryAtoms[c]);
                else
                    ret.AddMandatoryAtom(mandatoryAtoms[c]);
            }

            return ret;
        }

        /// <summary>
        /// Read in a genome from a file
        /// </summary>
        /// <param name="file"></param>
        /// <returns></returns>
        public static Genome Read(string file)
        {
            Genome ret = new Genome();

            ret.name = Path.GetFileNameWithoutExtension(file);

            BinaryReader read = new BinaryReader(new FileStream(file, FileMode.Open));

            int version = read.ReadInt32();

            int atomcount = read.ReadInt32();
            for (int c = 0; c < atomcount; c++) ret.AddAtom(Atom.Read(read));

            int mandatorycount = read.ReadInt32();
            for (int c = 0; c < mandatorycount; c++) ret.AddMandatoryAtom(Atom.Read(read));

            read.Close();
            return ret;
        }

        /// <summary>
        /// Serialize this genome to a file
        /// </summary>
        /// <param name="file"></param>
        public void Write(string file)
        {
            BinaryWriter write = new BinaryWriter(new FileStream(file, FileMode.Create));

            //File version
            write.Write(1);

            //Atoms
            write.Write(atoms.Count);
            for (int c = 0; c < atoms.Count; c++) atoms[c].Write(write);

            //Mandatories
            write.Write(mandatoryAtoms.Count);
            for (int c = 0; c < mandatoryAtoms.Count; c++) mandatoryAtoms[c].Write(write);

            write.Close();
        }
    }
}
